{-|
Module      : Data.Graph.Clustering.HLouvain
Description : Purely functional (Inductive) Louvain clustering
Copyright   : (c) Alexandre Delanoë, CNRS, 2020-Present
License     : AGPL + CECILL v3
Maintainer  : alexandre.delanoe+louvain@iscpif.fr
Stability   : experimental
Portability : POSIX

-}


module Data.Graph.Clustering.HLouvain 
  where

import Data.Ord (Down(..))
import Data.List (sortOn, foldl')
import Data.Graph.Inductive
-- import Data.Graph.Clustering.Louvain

------------------------------------------------------------------------
-- | Main Tools
------------------------------------------------------------------------

-- | Glustering (instead of clustering) since we are working on the glue
-- between nodes to create "glusters" of nodes
glustering :: DynGraph gr
           => gr a b
           -> [Module]
glustering g = foldl' (\a b -> modulare g b a) [] modules
  where
    modules :: [Module]
    modules = map (\n -> Module n [] (modularity' g [n])) ns

    ns :: [Node]
    ns = sortOn (Down . (deg g)) $ nodes g

------------------------------------------------------------------------
-- | Definitions
------------------------------------------------------------------------

data Module = Module { node    :: !Node
                     , modules :: ![Module]
                     , score   :: !Double
                     }
  deriving (Show)

-- | Divide en conquer method
modulare :: DynGraph gr
         => gr a b
         ->  Module
         -> [Module]
         -> [Module]
modulare _ m []  = [m]
modulare _ m [n] = sortOn (Down . score) [m,n]
modulare g m0@(Module n ms0 _) m1@(m@(Module n1 _ _):ms)
  | score1 > score2 = [Module n1 (modulare g m0 ms0) score1] ++ ms
  | otherwise       = [m]  ++ (modulare g m0 ms)
    where
      score1 = sum $ map (modularity g) [m0, m]
      score2 = sum $ map (modularity g) ([m0]++ms)

-- | need to flatten the Tree of modules to get a list of nodes
flat :: Module -> [Node]
flat (Module n [] _) = [n]
flat (Module n ns _) = [n] ++ (concat $ map flat ns)

-- | TODO Sum or Union -> QuickCheck
modularity :: DynGraph gr => gr a b -> Module -> Double
modularity g m = modularity' g (flat m)

modularity' :: DynGraph gr => gr a b -> [Node] -> Double
modularity' gr ns = coverage - edgeDensity
    where
        coverage :: Double
        coverage  = sizeSubGraph / sizeAllGraph
            where
                sizeSubGraph :: Double
                sizeSubGraph = fromIntegral ( size $ subgraph ns gr )

                sizeAllGraph :: Double
                sizeAllGraph = fromIntegral (size gr)

        edgeDensity :: Double
        edgeDensity = (sum (Prelude.map (\node -> (degree node) / links ) ns)) ** 2
            where
                degree :: Node -> Double
                degree node = fromIntegral (deg gr node)

                links :: Double
                links       = fromIntegral (2 * (size gr))

